# AUTOGENERATED by make_komauxitems from aux-items.py

AI_CONTENT_TYPE = 1 # (text)
AI_FAST_REPLY = 2 # (text)
AI_CROSS_REFERENCE = 3 # (text, conference, letterbox)
AI_NO_COMMENTS = 4 # (text)
AI_PERSONAL_COMMENT = 5 # (text)
AI_REQUEST_CONFIRMATION = 6 # (text)
AI_READ_CONFIRM = 7 # (text)
AI_REDIRECT = 8 # (conference, letterbox)
AI_X_FACE = 9 # (conference, letterbox, server)
AI_ALTERNATE_NAME = 10 # (text, conference, letterbox)
AI_PGP_SIGNATURE = 11 # (text)
AI_PGP_PUBLIC_KEY = 12 # (letterbox)
AI_E_MAIL_ADDRESS = 13 # (conference, letterbox, server)
AI_FAQ_TEXT = 14 # (conference, letterbox, server)
AI_CREATING_SOFTWARE = 15 # (text)
AI_MX_AUTHOR = 16 # (text)
AI_MX_FROM = 17 # (text)
AI_MX_REPLY_TO = 18 # (text)
AI_MX_TO = 19 # (text)
AI_MX_CC = 20 # (text)
AI_MX_DATE = 21 # (text)
AI_MX_MESSAGE_ID = 22 # (text)
AI_MX_IN_REPLY_TO = 23 # (text)
AI_MX_MISC = 24 # (text)
AI_MX_ALLOW_FILTER = 25 # (conference, letterbox)
AI_MX_REJECT_FORWARD = 26 # (conference, letterbox)
AI_NOTIFY_COMMENTS = 27 # (letterbox)
AI_FAQ_FOR_CONF = 28 # (text)
AI_RECOMMENDED_CONF = 29 # (server)
AI_ALLOWED_CONTENT_TYPE = 30 # (conference, letterbox, server)
AI_CANONICAL_NAME = 31 # (server)
AI_MX_LIST_NAME = 32 # (conference)
AI_SEND_COMMENTS_TO = 33 # (letterbox)
AI_WORLD_READABLE = 34 # (text)
AI_MX_REFUSE_IMPORT = 35 # (conference, letterbox)
AI_ELISP_CLIENT_READ_FAQ = 10000 # (letterbox)
AI_ELISP_CLIENT_REJECTED_RECOMMENDATION = 10001 # (letterbox)
AI_MX_MIME_BELONGS_TO = 10100 # (text)
AI_MX_MIME_PART_IN = 10101 # (text)
AI_MX_MIME_MISC = 10102 # (text)
AI_MX_ENVELOPE_SENDER = 10103 # (text)
AI_MX_MIME_FILE_NAME = 10104 # (text)
AI_MX_ALLOW_ENVELOPE_SENDER_REGEXP = 10105 # (conference, letterbox)
AI_KOMFEEDER_UNIQUE_ID = 10400 # (text)
AI_KOMFEEDER_URL = 10401 # (conference, text)
AI_KOMFEEDER_TITLE = 10402 # (conference)
AI_KOMFEEDER_AUTHOR = 10403 # (text)
AI_KOMFEEDER_DATE = 10404 # (text)

aux_item_number_to_name = {
  AI_CONTENT_TYPE: "content-type",
  AI_FAST_REPLY: "fast-reply",
  AI_CROSS_REFERENCE: "cross-reference",
  AI_NO_COMMENTS: "no-comments",
  AI_PERSONAL_COMMENT: "personal-comment",
  AI_REQUEST_CONFIRMATION: "request-confirmation",
  AI_READ_CONFIRM: "read-confirm",
  AI_REDIRECT: "redirect",
  AI_X_FACE: "x-face",
  AI_ALTERNATE_NAME: "alternate-name",
  AI_PGP_SIGNATURE: "pgp-signature",
  AI_PGP_PUBLIC_KEY: "pgp-public-key",
  AI_E_MAIL_ADDRESS: "e-mail-address",
  AI_FAQ_TEXT: "faq-text",
  AI_CREATING_SOFTWARE: "creating-software",
  AI_MX_AUTHOR: "mx-author",
  AI_MX_FROM: "mx-from",
  AI_MX_REPLY_TO: "mx-reply-to",
  AI_MX_TO: "mx-to",
  AI_MX_CC: "mx-cc",
  AI_MX_DATE: "mx-date",
  AI_MX_MESSAGE_ID: "mx-message-id",
  AI_MX_IN_REPLY_TO: "mx-in-reply-to",
  AI_MX_MISC: "mx-misc",
  AI_MX_ALLOW_FILTER: "mx-allow-filter",
  AI_MX_REJECT_FORWARD: "mx-reject-forward",
  AI_NOTIFY_COMMENTS: "notify-comments",
  AI_FAQ_FOR_CONF: "faq-for-conf",
  AI_RECOMMENDED_CONF: "recommended-conf",
  AI_ALLOWED_CONTENT_TYPE: "allowed-content-type",
  AI_CANONICAL_NAME: "canonical-name",
  AI_MX_LIST_NAME: "mx-list-name",
  AI_SEND_COMMENTS_TO: "send-comments-to",
  AI_WORLD_READABLE: "world-readable",
  AI_MX_REFUSE_IMPORT: "mx-refuse-import",
  AI_ELISP_CLIENT_READ_FAQ: "elisp-client-read-faq",
  AI_ELISP_CLIENT_REJECTED_RECOMMENDATION: "elisp-client-rejected-recommendation",
  AI_MX_MIME_BELONGS_TO: "mx-mime-belongs-to",
  AI_MX_MIME_PART_IN: "mx-mime-part-in",
  AI_MX_MIME_MISC: "mx-mime-misc",
  AI_MX_ENVELOPE_SENDER: "mx-envelope-sender",
  AI_MX_MIME_FILE_NAME: "mx-mime-file-name",
  AI_MX_ALLOW_ENVELOPE_SENDER_REGEXP: "mx-allow-envelope-sender-regexp",
  AI_KOMFEEDER_UNIQUE_ID: "komfeeder-unique-id",
  AI_KOMFEEDER_URL: "komfeeder-url",
  AI_KOMFEEDER_TITLE: "komfeeder-title",
  AI_KOMFEEDER_AUTHOR: "komfeeder-author",
  AI_KOMFEEDER_DATE: "komfeeder-date",
}
